package de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort

import de.fraunhofer.iem.dataprovider.dependency.dto.DependencyCreateDto
import de.fraunhofer.iem.dataprovider.dependency.dto.VulnerabilityCreateDto
import de.fraunhofer.iem.dataprovider.dependency.dto.VulnerabilityScoreCreateDto
import de.fraunhofer.iem.dataprovider.dependency.entity.DependencyEntity
import de.fraunhofer.iem.dataprovider.dependency.enumeration.VulnerabilityScoringSystemEnum
import de.fraunhofer.iem.dataprovider.dependency.service.DependencyService
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.taskManager.events.Event
import de.fraunhofer.iem.dataprovider.taskManager.tasks.Task
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.dto.OrtResultDto
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.json.AdvisorResultJson
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.json.OrtJson
import de.fraunhofer.iem.dataprovider.tool.dto.CreateToolDto
import de.fraunhofer.iem.dataprovider.tool.service.ToolService
import io.ktor.client.*
import io.ktor.client.call.*
import io.ktor.client.engine.cio.*
import io.ktor.client.plugins.contentnegotiation.*
import io.ktor.client.request.*
import io.ktor.client.statement.*
import io.ktor.serialization.kotlinx.json.*
import kotlinx.serialization.json.*
import java.util.*

class OrtApiTask(
    override val responseChannel: suspend (event: Event) -> Unit,
    val repoId: UUID,
    private val dependencyService: DependencyService,
    private val toolService: ToolService,
    private val repositoryService: RepositoryService
) : Task() {
    val json = Json { ignoreUnknownKeys = true }
    override suspend fun execute() {

        val client = HttpClient(CIO) {
            install(ContentNegotiation) {
                json(
                    json
                )
            }
        }

        val response: HttpResponse = client.get("http://localhost:3000/ort")
        val ortJson = response.body<OrtJson>()
        println(response.status)
        println(ortJson)
        client.close()
        val ortDto = ortJsonToDto(ortJson)
        val dependencies = mutableListOf<DependencyCreateDto>()

        ortDto.packages.forEach { p ->
            if (p.nameAndVersion != null) {
                dependencies.add(DependencyCreateDto(name = p.nameAndVersion))
            }
        }

        ortDto.advisorResults.forEach { advisorResult ->

            var dependency = dependencies.find { it.name == advisorResult.identifier }

            if (dependency == null) {
                dependency = DependencyCreateDto(advisorResult.identifier)
                dependencies.add(dependency)
            }
            advisorResult.vulnerabilities.forEach { vulnerability ->
                if (vulnerability.id != null) {

                    val vulnerabilityScoreDtos =
                        vulnerability.references.mapNotNull { ref ->
                            if (ref.severity != null && ref.scoringSystem != null) {
                                val scoringSystem = VulnerabilityScoringSystemEnum.fromString(ref.scoringSystem)
                                scoringSystem?.let {
                                    return@mapNotNull VulnerabilityScoreCreateDto(severity = ref.severity, it)
                                }
                            }
                            return@mapNotNull null
                        }
                    val vulnerabilityDto =
                        VulnerabilityCreateDto(vulnerability.id, vulnerabilityScoreDtos.toMutableList())
                    dependency.vulnerabilities.add(vulnerabilityDto)
                }
            }
        }

        storeResultsInDb(dependencies)
    }

    private fun ortJsonToDto(rawJson: OrtJson): OrtResultDto {
        val results = mutableListOf<AdvisorResultJson>()
        rawJson.advisor?.results?.advisorResults?.forEach { (key, value) ->
            if (value is JsonArray) {
                value.forEach { res ->
                    if (res is JsonObject) {

                        val flat = buildJsonObject {
                            res.entries.forEach { (eKey, eValue) ->
                                put(eKey, eValue)
                            }
                            put("identifier", key)
                        }

                        val advisorResult = json.decodeFromJsonElement<AdvisorResultJson>(flat)
                        results.add(advisorResult)
                    }
                }
            }
        }
        val packages = rawJson.analyzer?.result?.packages?.toList() ?: listOf()
        return OrtResultDto(advisorResults = results, packages = packages)
    }

    // TODO: Requires a database purge / association with the tool run
    private suspend fun storeResultsInDb(result: List<DependencyCreateDto>) {
        val dependencies = mutableListOf<DependencyEntity>()
        result.forEach { dependency ->
            val dependencyEntity = dependency.toDbObject()
            dependencyService.save(dependencyEntity)
            dependencies.add(dependencyEntity)
        }
        val tool = toolService.findOrCreateTool(CreateToolDto("ORT", "ORT", "docker-snapshot"))
        val repo = repositoryService.findRepoByID(repoId)
        if (repo != null) {
            repo.toolEntities.add(tool)
            repo.dependencyEntities.addAll(dependencies)
            repositoryService.save(repo)
        }

    }

}
